#include "linux/pport.h"

/* Simple access to the ioports of the parallel port. The
   driver allows two modes: in IOCTL mode the status and
   control registers are accessed using ioctl calls, in
   NON-IOCTL mode a simple protocol on the input data allows
   to access all registers using the read and write methods.
   This makes it possible for tools not able to use the ioctl
   call to access all registers.


DOSEQUENCE:
  The dosequence mode of operation can be used to do timing critical actions
  on the parallel port. A running sequence can't be chancelled.

  A Sequence consists of a series of 4byte ints, the first byte is the 
  action, the second a value, the last two bytes a delay in usec. If the
  value in the delay field is > 0, the process execution is delayed for that
  amount of usec !!! after !!! the command was executed.
  If the command is a read_value command, the value is written into the value
  field of the passed data.
 */


static struct parport *portlist = NULL;


static long long pport_lseek(struct file *file, loff_t offset, int origin);
static ssize_t pport_read(struct file *file, char *buffer, size_t count, loff_t *ppos);
static ssize_t pport_write(struct file *file, const char *buffer, size_t count, loff_t *ppos);
static int pport_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);
static int pport_open(struct inode *inode, struct file *file);
static int pport_release(struct inode * inode, struct file * file);


static struct file_operations pport_fops = {
#ifdef KERNEL_2_2
        pport_lseek,
        pport_read,
        pport_write,
        NULL,           /* lp_readdir */
        NULL,           /* lp_poll */
        pport_ioctl,
        NULL,           /* lp_mmap */
        pport_open,
        NULL,           /* flush */
        pport_release,
        NULL,           /* fsync */
        NULL,           /* fasync */
        NULL,           /* check media change */
        NULL,           /* revalidate */
        NULL            /* lock */ 
#else
#ifdef KERNEL_2_4
        owner:          THIS_MODULE,
	llseek:         pport_lseek,
        read:           pport_read,
	write:		pport_write,
        ioctl:          pport_ioctl,
        open:           pport_open,
        release:        pport_release,
#endif
#endif
};



// internal functions


void freeBuffers(struct parport *p) {
  if(p->indata) { kfree(p->indata); p->indata=NULL; }
  if(p->outdata) { kfree(p->outdata); p->outdata=NULL; }
}


void addOutDataByte(struct parport *p, int value) {
  p->outdata[p->outdatahead]=value;
  p->outdatahead++;
  if(p->outdatahead==p->outdatabuffersize) p->outdatahead=0;
  p->outdatalength++;
  if(p->outdatalength>p->outdatabuffersize) {
// outdata buffer overflow, write kernel warning once
    p->outdatalength=p->outdatabuffersize;
    if(!(p->flags&PPORT_OUT_OVERFLOW)) {
      printk(KERN_DEBUG "pport.248: output buffer overflow occured, data dropped: increase buffer sizes");
      p->flags|=PPORT_OUT_OVERFLOW;
    }
  }
}


void executeSequence(struct parport *p, int offset, int steps) {
  int instruction;
  int steppos;

#ifdef PPORT_DEBUG
  printk(KERN_DEBUG "pport.90: executing %d instructions at %d\n", steps, offset);
#endif

// execute the instructions

  for(steppos=0; steppos<steps; steppos++) {
    memcpy(&instruction, p->indata+offset+(steppos<<2), 4);

#ifdef PPORT_DEBUG
    printk(KERN_DEBUG "pport.218: executing instruction 0x%X\n", instruction);
#endif

    switch(instruction&0x3) {
      case PPORT_WBYTE:
        outb((instruction>>8)&0xFF, p->base);
        break;
      case PPORT_WSTATUS:
// clear the interrupt flag
        outb((instruction>>8)&0xEF, p->base+2);
        break;
      case PPORT_RBYTE:
        addOutDataByte(p, inb(p->base));
        break;
      case PPORT_RSTATUS:
        addOutDataByte(p, inb(p->base+1));
        break;
    }
    instruction=(instruction>>16)&0xFFFF;
    if(instruction>0) udelay(instruction);
  }
}



//***************************************************************************//
//									     //
//				file operation handlers			     //
//									     //
//***************************************************************************//


static long long pport_lseek(struct file *file, loff_t offset, int origin) {
  return -ESPIPE;
}


/** Main read method:
 *
 *  In ioctrl mode the method will return count bytes
 *  of data read from the data-register as fast as possible.
 *
 *  In non-ioctrl mode the driver will empty the outdata buffer
 *  and then stop.
 */
static ssize_t pport_read(struct file *file, char *buffer, size_t count, loff_t *ppos) {
  int i, j, k, fillpos;
  struct parport *p;

  if(count<0) return(-EINVAL);
  if(!count) return(0);

  p=(struct parport *)file->private_data;
#ifdef PPORT_DEBUG
  printk(KERN_DEBUG "pport.65: reading from port %lX.\n", p->base);
#endif

//  minor = MINOR(inode->i_rdev);

  fillpos=0;
  if(p->flags&PPORT_IOCTL_MODE) {
    while(fillpos!=count) {
// use the outdata buffer for fast reads and do the
// (slower) copy_to_user just once
      j=count-fillpos;
      j=j>p->outdatabuffersize?p->outdatabuffersize:j;
      for(k=0; k<j; k++) {
        p->outdata[k]=inb(p->base);
      }
// asm/uaccess.h: copy_to_user(*to, *from, n);
      copy_to_user(buffer+fillpos, p->outdata, j);
      fillpos+=j;
    }
  } else {
// just return the data from the circular databuffer
    if(p->outdatalength<count) count=p->outdatalength;

    while(fillpos!=count) {

// determine the number of missing bytes in the buffer that
// could be copied from the data buffer 
      i=count-fillpos;

// j points now to first readable pos
      j=p->outdatahead-p->outdatalength;
      if(j<0) j+=p->outdatabuffersize;

// k is max copyable bytes without circular buffer loop
      k=p->outdatabuffersize-j;
      if(k>i) k=i;

      copy_to_user(buffer+fillpos, p->outdata+j, k);
      fillpos+=k;
      p->outdatalength-=k;
    }
  }

  return(count);
}




/** Main driver write implementation: the driver may work in two
 *  different modes:
 *
 *  In ioctl mode, the write data is written
 *  sequentially to the data output register. The high CPU
 *  speed will produce rapidly changing signals at the
 *  parallel port if more than one byte is written per call.
 *  Some port output drivers will be to slow to follow.
 *
 *  In normal non-ioctl mode, the driver will collect the written data
 *  bytes in the parport's private input buffer until enough
 *  data is available for execution. Normal read/write commands
 *  can't produce overflows, they are executed as they come.
 *  Long sequences may result in overflows, the driver tries
 *  to handle that neatly by executing the already available
 *  part of the sequence (with the correct timings) and then
 *  add a pseudo sequence header containing a reduced number
 *  of steps so that the remaining steps are read.
 */
static ssize_t pport_write(struct file *file, const char *buffer, size_t count, loff_t *ppos) {
  int i, j;
  struct parport *p;
  int readpos, parsepos, lastparsepos;
  int sequencesteps;

  if(count<0) return(-EINVAL);
  if(!count) return 0;

  p=(struct parport *) file->private_data;
//  minor = MINOR(inode->i_rdev);

  if(p->flags&PPORT_IOCTL_MODE) {
// just write one byte to the base reg, use the 
    readpos=0;
    while(readpos<count) {
      i=count-readpos;
      i=i>p->indatabuffersize?p->indatabuffersize:i;
      copy_from_user(p->indata, buffer+readpos, i);
      for(j=0; j<i; j++) {
        outb(p->indata[j], p->base);
      }
      readpos+=i;
    }

    return(readpos);
  }


#ifdef PPORT_DEBUG
  printk(KERN_DEBUG "pport.119: writing %d bytes to port 0x%lX.\n", count,
         p->base);
#endif

  readpos=0;
  while(readpos<count) {
    i=count-readpos;
    j=p->indatabuffersize-p->indatapos;
    if(i>j) i=j;
// i now stores max copyable number of bytes

#ifdef PPORT_DEBUG
    printk(KERN_DEBUG "pport.132: filling buffer from %d with %d bytes.\n",
           p->indatapos, i);
#endif

    copy_from_user(p->indata+p->indatapos, buffer+readpos, i);
    readpos+=i;
    p->indatapos+=i;

    for(parsepos=0; parsepos<p->indatapos; ) {
      lastparsepos=parsepos;
 
      switch(p->indata[parsepos]&PPORT_ACTION) {

        case PPORT_WBYTE:
          if((parsepos+2)>p->indatapos) break;
#ifdef PPORT_DEBUG
          printk(KERN_DEBUG "pport.286: writing 0x%X.\n",
                 p->indata[parsepos+1]&0xFF);
#endif
          outb(p->indata[parsepos+1], p->base);
          parsepos+=2;
          break;
        case PPORT_WSTATUS:
          if((parsepos+2)>p->indatapos) break;
#ifdef PPORT_DEBUG
          printk(KERN_DEBUG "pport.295: setting ctrl-reg to %X.\n",
                 p->indata[parsepos+1]&0xFF);
#endif
// clear the interrupt flag
          outb(p->indata[parsepos+1]&0xEF, p->base+2);
          parsepos+=2;
          break;
        case PPORT_RBYTE:
          addOutDataByte(p, inb(p->base));
          parsepos++;
          break;

        case PPORT_RSTATUS:
          i=inb(p->base+1);
#ifdef PPORT_DEBUG
          printk(KERN_DEBUG "pport.170: read status %X.\n", i&0xFF);
#endif
          addOutDataByte(p, i);
          parsepos++;
          break;

        case PPORT_SEQUENCE:
// sequence header incomplete, leave loop to get more data;
          if((parsepos+5)>p->indatapos) break;

          memcpy(&sequencesteps, p->indata+parsepos+1, 4);
#ifdef PPORT_DEBUG
          printk(KERN_DEBUG "pport.184: executing sequence at %d, steps %d.\n",
                 parsepos, sequencesteps);
#endif

          if(sequencesteps>((p->indatapos-parsepos-5)>>2)) {
// sequence not complete

            if((parsepos!=0)||(p->indatapos!=p->indatabuffersize)) {
// there is still space in the input buffer, wait for user
// to fill it up with missing bytes
              break;
            }

// buffer is filled completely with sequence, but it still
// does not fit into it: just execute as much as possible
// and update the sequencesteps field in the input buffer
            printk(KERN_WARNING "Input buffer overflow, sequence timing may be altered.\n");

// just execute as much as possible, i is number of steps
            i=(p->indatapos-5)>>2;
            executeSequence(p, 5, i);

            j=5+(i<<2);
            memmove(p->indata+5,   // to
                    p->indata+j,  // from
                    p->indatabuffersize-j); // #bytes

            p->indatapos=5+p->indatabuffersize-j;
            parsepos=0;

            sequencesteps-=i;
            memcpy(p->indata+1, &sequencesteps, 4);
          } else {
// normal execution of sequence
            executeSequence(p, parsepos+5, sequencesteps);
            parsepos+=5+(sequencesteps<<2);
          }

          break;
        default:
          printk(KERN_WARNING "Unparseable command %X.\n",
                 p->indata[parsepos]&0xFF);
          parsepos++;
      }

      if(parsepos==lastparsepos) {
// nothing read, end the loop and see if there is more to read
// from the user space
        break;
      }
    }

// parsepos is the number of bytes parsed so far
#ifdef PPORT_DEBUG
    printk(KERN_DEBUG "pport.256: leaving data from %d to %d in buffer, output buffer fill %d.\n", parsepos, p->indatapos, p->outdatalength);
#endif
    if(parsepos==p->indatapos) p->indatapos=0;
    else if(parsepos!=0) {
      memmove(p->indata, p->indata+parsepos, p->indatapos-parsepos);
      p->indatapos-=parsepos;
    }
  }

  return(readpos);
}





/* Do the ioctrl operation
 */
static int pport_ioctl(struct inode *inode, struct file *file,
                       unsigned int cmd, unsigned long arg) {
  unsigned int minor = MINOR(inode->i_rdev);
  int i, n, len;

#ifdef PPORT_DEBUG
  printk(KERN_DEBUG "pport%d ioctl cmd %X with 0x%X\n", minor, cmd, (int)arg);
#endif

  if (minor >= PPORT_MAX_PORTS) return -ENODEV;
  if (!(portlist[minor].flags & PPORT_EXISTS)) return -ENODEV;

  if((MAJOR(inode->i_rdev)!=PPORT_MAJOR)||((cmd>>8)!=PPORT_MAJOR)) {
#ifdef PPORT_DEBUG
  printk(KERN_DEBUG "pport%d: major number missmatch in ioctl cmd %X\n",
             minor, cmd);
#endif
// major mismatch
    return(-ENODEV);
  }


  switch (cmd&PPORT_IOCTL_CMD) {
    case PPORT_IOCTL_MODEOFF:
      portlist[minor].flags&=(~PPORT_IOCTL_MODE);
      break;
    case PPORT_IOCTL_MODEON:
      portlist[minor].flags|=PPORT_IOCTL_MODE;
      break;
    case PPORT_WRITESTATUS:
// clear the interrupt flag
      outb(arg&0xEF, portlist[minor].base+2);
      break;
    case PPORT_READSTATUS:
      i=inb(portlist[minor].base+1);
      copy_to_user((long*)arg, &i, 1);
      break;
    case PPORT_DOSEQUENCE:
      copy_from_user((long*)arg, &len, 4);  // 
      len++;
      for(n=1; n<len; n++) {
        copy_from_user(((long*)arg)+n, &i, 4);
        switch(i&0x3) {
          case PPORT_WBYTE:
            outb((i>>8)&0xFF, portlist[minor].base);
            break;
          case PPORT_WSTATUS:
// clear the interrupt flag
            outb((i>>8)&0xEF, portlist[minor].base+2);
            break;
          case PPORT_RBYTE:
            i&=0xFFFF00FF;
            i|=((inb(portlist[minor].base)&0xFF)<<8);
            copy_to_user(((long*)arg)+n, &i, 4);
            break;
          case PPORT_RSTATUS:
            i&=0xFFFF00FF;
            i|=((inb(portlist[minor].base+1)&0xFF)<<8);
            copy_to_user(((long*)arg)+n, &i, 4);
            break;
        }
        i=(i>>16)&0xFFFF;
        if(i>0) udelay(i);
      }
      break;
    default:
      return(-EINVAL);
  }
  return(0);
}




static int pport_open(struct inode *inode, struct file *file) {
  unsigned int minor = MINOR(inode->i_rdev);

  if(minor >= PPORT_MAX_PORTS) {
#ifdef PPORT_DEBUG
    printk(KERN_DEBUG "pport.187: minor %d greater than maxdevs.\n", minor);
#endif
    return -ENXIO;
  }

  if((portlist[minor].flags & PPORT_EXISTS)==0) {
#ifdef PPORT_DEBUG
    printk(KERN_DEBUG "pport.193: minor %d device doesn't exist.\n", minor);
#endif
    return -ENXIO;
  }

  if(portlist[minor].flags & PPORT_IS_OPEN) {
#ifdef PPORT_DEBUG
    printk(KERN_DEBUG "pport.201: minor %d device is busy.\n", minor);
#endif
    return -EBUSY;

  }

//  portlist[minor].flags|=PPORT_IS_OPEN;
  file->private_data = portlist+minor;
  MOD_INC_USE_COUNT;
  return(0);
}


static int pport_release(struct inode * inode, struct file * file) {
  unsigned int minor = MINOR(inode->i_rdev);

//  pport_reset(minor);
  if (portlist[minor].flags&PPORT_IS_OPEN) {
    portlist[minor].flags&=(~PPORT_IS_OPEN);
  }

  file->private_data=NULL;
  MOD_DEC_USE_COUNT;
  return(0);
}




static int pport_probe(int offset) {
  int base;

  base = portlist[offset].base;
#ifdef PPORT_DEBUG
  printk(KERN_DEBUG "probing pport minor %d at 0x%X\n", offset, base);
#endif

  if(base < 0) return(-1);     /* de-configured by command line */

  if(check_region(base, 3) < 0) {
#ifdef PPORT_DEBUG
    printk(KERN_DEBUG "pport.340: region at 0x%X demanded by other driver\n", base);
#endif
    return(-1);
  }

  portlist[offset].flags|=PPORT_EXISTS;
  return(0);
}









#ifdef MODULE
#define pport_init init_module

static int io[PPORT_MAX_PORTS] = {0, };

MODULE_PARM(io, "1-" __MODULE_STRING(PPORT_MAX_PORTS) "i");

#endif

int pport_init(void) {
  int offset = 0;
  int count = 0;
  int i=0;
#ifdef MODULE
  int failed = 0;
#endif


#ifdef PPORT_DEBUG
  printk(KERN_DEBUG "pport.279: initializing pport driver.\n");
#endif

  if (register_chrdev(PPORT_MAJOR, "pport", &pport_fops)) {
    printk(KERN_WARNING "pport: unable to get major %d\n", PPORT_MAJOR);
    return -EIO;
  }

#ifdef PPORT_DEBUG
  printk(KERN_DEBUG "pport.288: registered char device pport for major %d.\n",
                    PPORT_MAJOR);
#endif


#ifdef MODULE

#ifdef PPORT_DEBUG
  printk(KERN_DEBUG "pport.395: scanning io values %lX.\n", (long)io);
#endif

  if(!io) {
    printk(KERN_WARNING "pport.297: io params is null reference.\n");
    unregister_chrdev(PPORT_MAJOR, "pport");
    return -EIO;
  }


/* When user feeds parameters, use them */
  portlist=(struct parport*)kmalloc(PPORT_MAX_PORTS*sizeof(struct parport), GFP_KERNEL);

  if(!portlist) {
    printk(KERN_WARNING "pport.308:  no memory for parport structures.\n");
    unregister_chrdev(PPORT_MAJOR, "pport");
    return -ENOMEM;
  }

  for (offset=0; offset < PPORT_MAX_PORTS; offset++) {
    int specified=0;

    if (io[offset] != 0) {
#ifdef PPORT_DEBUG
  printk(KERN_DEBUG "pport.419: reconfig of pport minor %d at %lX with %X\n",
                     offset, portlist[offset].base, io[offset]);
#endif
      portlist[offset].base = io[offset];
      portlist[offset].flags=0;

// input buffer preparation
      portlist[offset].indatapos=0;
      portlist[offset].indatabuffersize=1<<10;
      portlist[offset].indata=
                 kmalloc(portlist[offset].indatabuffersize, GFP_KERNEL);

// prepare the output buffer
// a good size estimate is 1/4 of input buffer size: then the
// inputbuffer could be filled completely with a sequence of read
// commands without overflowing the outdata buffer
      portlist[offset].outdatabuffersize=portlist[offset].indatabuffersize>>2;
      portlist[offset].outdatahead=0;
      portlist[offset].outdatalength=0;
      portlist[offset].outdata=
                 kmalloc(portlist[offset].outdatabuffersize, GFP_KERNEL);
      specified++;
    } else {
      portlist[offset].base=-1; 
      portlist[offset].indata=NULL;
      portlist[offset].outdata=NULL;
    }

    if(pport_probe(offset) < 0) {
#ifdef PPORT_DEBUG
      printk(KERN_DEBUG "pport.433: probe for pport%d at 0x%X failed.\n",
             offset, (int)portlist[offset].base);
#endif
      printk(KERN_INFO "pport%d: Probe failed\n", offset);
      failed++;
      freeBuffers(&(portlist[offset]));
    } else {
      sprintf(portlist[offset].devname, "pport%d", offset);

// try to reserve the region
// see /usr/src/linux-2.4/include/linux/ioport.h
//      i=request_region(portlist[offset].base, 3, &(portlist[offset].devname[0]));
      i=request_region(portlist[offset].base, 3, portlist[offset].devname);
      if(i!=0) {
        printk(KERN_INFO "pport%d: request region returned %d\n", offset, i);
      }

      count++;
      printk(KERN_INFO "pport%d: Registered\n", offset);

// clear the interrupt flag
      outb(inb(portlist[offset].base+2)&0xEF, portlist[offset].base+2);
    }
  }

        /* Successful specified devices increase count
         * Unsuccessful specified devices increase failed
         */

  if(count) {
    printk(KERN_INFO "pport.338: initialized %d devices.\n", count);
    return 0;
  }

  if (failed) {
    printk(KERN_WARNING "pport.315: No override devices found.\n");
    unregister_chrdev(PPORT_MAJOR, "pport");
    return -EIO;
  }


#endif
  if (count == 0) printk(KERN_WARNING "pport: Driver configured but no interfaces found.\n");
  return 0;
}






#ifdef MODULE
void cleanup_module(void) {
  int pcnt;

  unregister_chrdev(PPORT_MAJOR, "pport");
  if(!portlist) return;

  for (pcnt = 0; pcnt<PPORT_MAX_PORTS; pcnt++) {
    if(portlist[pcnt].flags&PPORT_EXISTS) {
      freeBuffers(&(portlist[pcnt]));
      release_region(portlist[pcnt].base, 3);
    }
  }
  kfree(portlist);
}
#endif




